﻿using System;
using System.Net;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using Microsoft.Phone.Controls;
using System.Windows.Navigation;
using Microsoft.Phone.Shell;
using System.Collections.Generic;

namespace LWP.NavigationSamples
{
    public class BackNavigationHelper
    {
        private PhoneApplicationFrame _frame;
        private Uri _backNavigationUri;

        static BackNavigationHelper()
        {
            PhoneApplicationService.Current.Deactivated +=new EventHandler<DeactivatedEventArgs>(OnPhoneActivationServiceDeactivated);
            PhoneApplicationService.Current.Activated +=new EventHandler<ActivatedEventArgs>(OnPhoneApplicationServiceActivated);
            PhoneApplicationService.Current.Launching +=new EventHandler<LaunchingEventArgs>(OnPhoneApplicationServiceLaunching);
        } 

        public event EventHandler<BackNavigationStartedEventArgs> BackNavigationStarted;
        public event EventHandler<BackNavigationCompletedEventArgs> BackNavigationCompleted;

        private static BackNavigationHelper _instance; 

        public static BackNavigationHelper Instance
        {
            get
            {
                if (_instance == null)
                {
                    _instance = new BackNavigationHelper();
                }
                return _instance; 

            } 
        }

        public bool IsInitialized
        {
            get;
            private set;
        }

        public bool IsLoaded
        {
            get;
            private set;
        } 

        public bool Initialize(PhoneApplicationFrame frame , bool autoJournal , bool hideFrameOnNav, bool validateUriBeforeNav )
        {
            if (!IsInitialized)
            {
                frame.Navigated += new System.Windows.Navigation.NavigatedEventHandler(frame_Navigated);
                frame.Navigating += new System.Windows.Navigation.NavigatingCancelEventHandler(frame_Navigating);
                frame.NavigationStopped += new NavigationStoppedEventHandler(frame_NavigationStopped);                
                _frame = frame;
                this.AutoJournal = autoJournal ;
                this.HideFrameonBackNav = hideFrameOnNav ; 
                this.ValidateUriBeforeBackNav = validateUriBeforeNav; 
                IsInitialized = true;
                
                return true; 
            }             
            return false; 
        }


        #region State 
        const string BackNavHistoryKey = "BackNavigationHelper.History";
        const string HideFrameOnBackNavKey = "BackNavigationHelper.HideFrameOnBackNav";
        const string AutoJournalKey = "BackNavigationHelper.AutoJournal";
        const string ValidateUriBeforeBackNavKey = "BackNavigationHelper.ValidateUriBeforeBackNav"; 

        static void OnPhoneApplicationServiceLaunching(object sender, LaunchingEventArgs e)
        {
            
          //   Instance.Load(); 
   
        }


     
        void SaveState (string key, object value)
        {
            PhoneApplicationService.Current.State[key] = value; 
        } 

        static void OnPhoneActivationServiceDeactivated(object sender, DeactivatedEventArgs e)
        {
            Instance.Save(); 
        }

        void Save( )
        {
            SaveState(BackNavHistoryKey, journalingStack.ToArray());
            SaveState(HideFrameOnBackNavKey, HideFrameonBackNav);
            SaveState(ValidateUriBeforeBackNavKey, ValidateUriBeforeBackNav);
            SaveState(AutoJournalKey, AutoJournal); 
        }

        bool isActivating = false; 
        void Load()
        {

            /* TODO: Right now we allow duplication of Loading when a chooser that does not 
             tombstone but does deactivate  
             */ 
            System.Diagnostics.Debug.Assert(
                ( IsLoaded == true && PhoneApplicationService.Current.StartupMode == StartupMode.Activate) || 
                  IsLoaded == false ); 

            if ( PhoneApplicationService.Current.StartupMode == StartupMode.Activate)
            {
                
                System.Diagnostics.Debug.WriteLine("Activated"); 
                Uri[] historyentries = LoadState<Uri[]>(BackNavHistoryKey, null );
                if (historyentries != null)
                {
                    journalingStack.Clear(); 
                    for ( int x = historyentries.Length-1; x >= 0; x-- ) 
                        journalingStack.Push(historyentries[x]);
                } 

                Instance.HideFrameonBackNav = LoadState<bool>(HideFrameOnBackNavKey, true);
                Instance.AutoJournal = LoadState<bool>(AutoJournalKey, true);
                Instance.ValidateUriBeforeBackNav = LoadState<bool>(ValidateUriBeforeBackNavKey, false);
            } 
#if DEBUG 
            else
                System.Diagnostics.Debug.WriteLine("Launched"); 
#endif 
            this.IsLoaded = true; 
        }

        T LoadState<T>(string key, T defaultValue )
        {
            if (PhoneApplicationService.Current.State.ContainsKey(key))
            {
                return (T)PhoneApplicationService.Current.State[key]; 
            }
            return defaultValue; 

        }

        private bool IsTombstoned { get ; set ;}   

        static void OnPhoneApplicationServiceActivated(object sender, ActivatedEventArgs e)
        {
            Instance.IsTombstoned = true; 

            //TODO:   This checks if this was truly a tombstone.. 
            //  Some choosers do not tombstone the app in full..  they do deactivate but 
            if ( PhoneApplicationService.Current.StartupMode == StartupMode.Activate ) 
                Instance.Load(); 
        }

        #endregion 

         
        

        bool _autoJournal; 
        public bool AutoJournal
        {
            get
            {
                return _autoJournal; 
            }
            set
            {
                if (_autoJournal != value)
                {
                    _autoJournal = value;

                    if (!AutoJournal)
                        StopJournaling(); 
                } 
            } 
        } 
        void frame_NavigationStopped(object sender, NavigationEventArgs e)
        {
            if (IsDoingBackNavigation)
                StopBackNavigation( false, true  );

            IsNavigationInProgress = false; 
        }

        public bool IsDoingBackNavigation
        {
            get;
            private set;
        }

        System.Collections.Generic.Stack<Uri> journalingStack = new System.Collections.Generic.Stack<Uri>(); 
        void StopJournaling()
        {             
            journalingStack.Clear(); 
        }
 
        public void StopBackNavigation(bool completed, bool userInitiated)
        {
            OnStopBackNavigation( completed, userInitiated); 
        }

        public virtual void OnStopBackNavigation(bool completed, bool userInitiated) 
        { 
          
            EventHandler<BackNavigationCompletedEventArgs> eh = this.BackNavigationCompleted ;
            BackNavigationCompletedEventArgs args = new BackNavigationCompletedEventArgs { Destination = _backNavigationUri, NavigationCompleted = completed, UserCancelled = userInitiated };
            if (eh != null)
                eh(this, args);

             IsDoingBackNavigation = false ; 
            _backNavigationUri = null ;

            if (HideFrameonBackNav)
                _frame.Opacity = frameOpacity;

#if DEBUG 
            System.GC.Collect(); 
            System.GC.Collect(); 
            System.GC.Collect(); 
#endif 

        }

        public Uri UriInProgress
        {
            get;
            internal set;
        } 

        public bool IsNavigationInProgress
        {
            get; 
            private set;
        } 
        double frameOpacity; 


        public void StartBackNavigation(Uri targetUri)
        {
            OnStartBackNavigation(targetUri);  
        }

public virtual bool OnStartBackNavigation(Uri targetUri)
{
    if (AutoJournal && ValidateUriBeforeBackNav)
    {
        if (!journalingStack.Contains(targetUri))
            return false;                     
    } 

    EventHandler<BackNavigationStartedEventArgs> start = this.BackNavigationStarted;
    BackNavigationStartedEventArgs args = new BackNavigationStartedEventArgs { Destination = targetUri }; 
    if (start != null)
        start(this, args );

    // if no page objected to a back nav then do it. 
    if (!args.Cancel)
    {
        IsDoingBackNavigation = true;
        _backNavigationUri = targetUri;

        if (HideFrameonBackNav)
        {
            frameOpacity = _frame.Opacity; 
            _frame.Opacity = 0;
        } 

        _frame.GoBack();             
    } 

    return !args.Cancel; 
}

        NavigationMode currentDirection; 

        void frame_Navigating(object sender, System.Windows.Navigation.NavigatingCancelEventArgs e)        
        {
            
            currentDirection = e.NavigationMode;

            UriInProgress = e.Uri; 
#if DEBUG 
            switch (e.NavigationMode)
            {
                case NavigationMode.Back:                    
                    break; 
                case NavigationMode.New:
                    break; 
                case NavigationMode.Refresh:
                    System.Diagnostics.Debug.Assert(false, "trying to learn when this happens"); 
                    break;
                case NavigationMode.Forward:
                    System.Diagnostics.Debug.Assert(false, "trying to learn when this happens"); 
                    break; 
            } 
#endif 

            this.IsNavigationInProgress = true; 
        }


        public void SafeNavigate(Uri targetUri, bool preventCircles = false )
        {
            if (IsNavigationInProgress)
            {
                if (UriInProgress != targetUri)
                {
                    _frame.StopLoading();
                    System.Diagnostics.Debug.WriteLine( "Stopped Loading a nav");
                }
                else
                {
                    System.Diagnostics.Debug.WriteLine("Ignore a double click ");
                    return;
                }
            }
            else if (targetUri == _frame.Source)
            {
                System.Diagnostics.Debug.WriteLine("Ignore fragment navigation");
                return; 
            }


            if (preventCircles)
            {
                if (journalingStack.Contains(targetUri))
                    this.StartBackNavigation(targetUri); 

            } 
            else 
                _frame.Navigate(targetUri);
        } 

void frame_Navigated(object sender, System.Windows.Navigation.NavigationEventArgs e)
{
    System.Diagnostics.Debug.WriteLine("Navigated to" + e.Uri); 
    if (AutoJournal)
    {
        if (currentDirection == NavigationMode.Back)
        {
            if (journalingStack.Count > 0)
            {

                if (!IsTombstoned)
                {
                    journalingStack.Pop();                            
                } 
                else
                    IsTombstoned = false; 
#if DEBUG
                if (journalingStack.Count > 0)
                {
                    Uri top = journalingStack.Peek();
                    System.Diagnostics.Debug.Assert(top == e.Uri);
                }
#endif
            }
            // we have reached pages that were not on our stack. This likely won't happen 
            // for real apps, but happens for demos when you turn journaling on/off.. 
            else
            {
                journalingStack.Push(e.Uri); 
            } 
        }

        if (currentDirection == NavigationMode.New)
        {
                // SKIP the tombstoning URLs 
            if ( e.Uri == new Uri ( "app://external", UriKind.RelativeOrAbsolute ) ) 
            {
                IsTombstoned = true; 
                System.Diagnostics.Debug.WriteLine ("*** External *** "); 
            }                   
            else

                journalingStack.Push(e.Uri); 
        } 
    }

    if (IsDoingBackNavigation )  
    {
        if (e.Uri != _backNavigationUri && _frame.CanGoBack)
        {                 
            _frame.GoBack();
            return;
        }
        else
        {
            StopBackNavigation(true, false);
        } 
                    
    } 
    IsNavigationInProgress = false; 
}

        public bool HideFrameonBackNav { get; set; }
        public bool ValidateUriBeforeBackNav { get; set; } 
         

        public static Settings DemoSettings = new Settings(); 

        public class  Settings
        {
            public bool HideAppBarOnNavigate { get; set; }
            public bool PreventLoopCreation { get; set; } 
        } 
    }


    public class BackNavigationStartedEventArgs : EventArgs
    {
        public Uri Destination;
        public bool Cancel; 
    }

    public class BackNavigationCompletedEventArgs : EventArgs
    {
        public Uri Destination;
        public bool NavigationCompleted;
        public bool UserCancelled; 
    }
}
